searchData={"items":[{"type":"module","title":"Decorum","doc":"Documentation for `Decorum`.","ref":"Decorum.html"},{"type":"function","title":"Decorum.and_then/2","doc":"Use this function to chain generators together when a generator is based on the\nvalue emitted by another generator.\n\nIn StreamData this funciton is called `bind`.\n\n`fun` is a function that takes a value from the given generator and\nreturns a generator.","ref":"Decorum.html#and_then/2"},{"type":"function","title":"Decorum.atom/1","doc":"Generates alphanumeric atoms.","ref":"Decorum.html#atom/1"},{"type":"function","title":"Decorum.binary/1","doc":"Generates binaries.\n\n`max_length` is used to cap the length of the binary. It defaults to `:none` and is the same as in `list_of/2`.","ref":"Decorum.html#binary/1"},{"type":"function","title":"Decorum.byte/0","doc":"Generates bytes.\n\nA byte is an integer between `0` and `255`.","ref":"Decorum.html#byte/0"},{"type":"function","title":"Decorum.check/1","doc":"","ref":"Decorum.html#check/1"},{"type":"function","title":"Decorum.check/2","doc":"","ref":"Decorum.html#check/2"},{"type":"function","title":"Decorum.check_all/2","doc":"`check_all` takes a Decorum struct and runs `test_fn`\nagainst the generated values.\n\nThis funciton will expand and eventually be called by a macro, but for now it's part\nof bootsrtapping the property testing functionality.\n\n`test_fn` should behaive like a normal ExUnit test. It throws an error if a test fails.\nUse the assert or other test helper macros inside that function.\n\nTODO: Create a Decorum.Error and raise that instead of ExUnit.AssertionError.","ref":"Decorum.html#check_all/2"},{"type":"function","title":"Decorum.constant/1","doc":"Create a generator that is not random and always returns the same value.","ref":"Decorum.html#constant/1"},{"type":"function","title":"Decorum.filter/3","doc":"Similar to Enum.filter/2\n\nFilters the generator. Returns only values for which fun returns a truthy value.\n\nUse `limit` to specify how many times the generator should be called before raising an error.","ref":"Decorum.html#filter/3"},{"type":"function","title":"Decorum.integer/1","doc":"Generates integers in the given range.\n\nRange handling borrowed from StreamData\n\nShrinks toward zero within the range.","ref":"Decorum.html#integer/1"},{"type":"function","title":"Decorum.list_of/2","doc":"Generates a list of values produced by the given generator.\n\n`max_length` is used to cap the length of the list. It defaults to `:none`.\n\nUses a biased coin flip to determine if another value should be gerenated\nor the list should be terminated.\n\nWhen `max_length` is a value other than `:none`,\nthen the probablility of terminating the list increases as the list size approaches `max_length`.\n\nWhen `max_length` is `:none` the probablility of generating another value is around 7/8.","ref":"Decorum.html#list_of/2"},{"type":"function","title":"Decorum.list_of_length/2","doc":"","ref":"Decorum.html#list_of_length/2"},{"type":"function","title":"Decorum.map/2","doc":"Similar to Enum.map/2\n\nReturns a generator where each element is the result of invoking fun\non each corresponding element of the given generator.","ref":"Decorum.html#map/2"},{"type":"function","title":"Decorum.new/1","doc":"Helper for creating Decorum structs from a generator function.","ref":"Decorum.html#new/1"},{"type":"function","title":"Decorum.one_of/1","doc":"Randomly selects one of the given generators.\n\n`generators` must be a list.","ref":"Decorum.html#one_of/1"},{"type":"function","title":"Decorum.prng_values/0","doc":"Creates a simple generator that outputs the values it gets from the prng.\n\nValues will be 32-bit positive integers.","ref":"Decorum.html#prng_values/0"},{"type":"function","title":"Decorum.stream/2","doc":"Used to run a Decorum generator with a specific Prng struct.\n\nTakes a Decorum struct and a Prng struct and returns a lazy Enumerable\nof generated values.","ref":"Decorum.html#stream/2"},{"type":"function","title":"Decorum.uniform_integer/1","doc":"Generates an integer between 0 and max.\n\nCurrently only supports 32-bit values and is not truly uniform as the use of `rem`\nhas a bias towards producing smaller numbers.","ref":"Decorum.html#uniform_integer/1"},{"type":"function","title":"Decorum.zip/1","doc":"Similar to Enum.zip/1\n\nZips corresponding elements from a finite collection of generators into a\ngenerator of tuples.","ref":"Decorum.html#zip/1"},{"type":"function","title":"Decorum.zip/2","doc":"Similar to Enum.zip/2\n\nZips corresponding elements from two generators into a generator of tuples.","ref":"Decorum.html#zip/2"},{"type":"type","title":"Decorum.generator_fun/1","doc":"","ref":"Decorum.html#t:generator_fun/1"},{"type":"type","title":"Decorum.max_length/0","doc":"","ref":"Decorum.html#t:max_length/0"},{"type":"type","title":"Decorum.t/1","doc":"","ref":"Decorum.html#t:t/1"},{"type":"type","title":"Decorum.value/0","doc":"","ref":"Decorum.html#t:value/0"},{"type":"module","title":"Decorum.Prng","doc":"`Prng` (pseudo random number generator) is a wrapper around the `:rand` module.\n\nIt has 2 states:\n1. Random\n  Uses `:rand` to generate random numbers and stores each one in history.\n\n2. Hardcoded\n  Used to replay a previously recorded (or simplified) history.\n\nWe represent random values using 32-bit non-negative integers because\nthey are easier to work with when shrinking the history.","ref":"Decorum.Prng.html"},{"type":"function","title":"Decorum.Prng.next!/1","doc":"Takes a `Prng` struct and returns a tuple with the next random value\nand an updated `Prng` struct.\n\nWhen in `Random` state, `next!/1` is not expected to fail.\n\nWhen in `Hardcoded` state, `next!/1` could raise a `Decorum.EmptyHistoryError`.\n\nGenerators will call `next!/1` to get a value to use when\ngenerating test values. They should also return the updated `Prng` struct.","ref":"Decorum.Prng.html#next!/1"},{"type":"opaque","title":"Decorum.Prng.t/0","doc":"","ref":"Decorum.Prng.html#t:t/0"},{"type":"extras","title":"Decorum","doc":"# Decorum\n\nProperty-based testing for Elixir with **shrinking that just works.**\n\nThis is an implementation of internal shrinking, which was first invented for the [Hypothesis](https://github.com/HypothesisWorks/hypothesis) library in Python.\n\nShrinking is the killer feature of property-based testing. It is the process of reducing a randomly generated failing test case into a small human-readable example.\nOther approaches to shrinking require code to be explicitly written to shrink each specific type of generated data.\nThe approach we are using is built on the concept that the series of random numbers that feeds the data generators can be simplified in a unified manner that will lead to simpler test cases for all data generators automatically.\n\nThere are specific implementations for some generators (such as `float` which is on the TODO list for this library) that are fine tuned for this method of shrinking.\nHowever, new generators built by composing the generators provided here should shrink well without additional thought or effort.\nIf you find an example of a generator and a test case that does not shrink well, please submit an issue so we can look into it.","ref":"readme.html"},{"type":"extras","title":"Status - Decorum","doc":"I'm currently in proof-of-concept mode. I've borowed some concepts from [StreamData](https://github.com/whatyouhide/stream_data) which is the de facto standard Elixir library for property-based testing. I've also leaned heavily on the [Elm test](https://github.com/elm-explorations/test) implementation.\n\nThere are no macros yet, so property tests are just regular ExUnit tests with names starting with \"property\" by convention.\n\nSee [Implementation Notes](NOTES.md) for my ongoing thoughts, questions, and TODOs.","ref":"readme.html#status"},{"type":"extras","title":"Background - Decorum","doc":"- https://youtu.be/WE5bmt0zBxg?si=fyg6R_O3iRrbuc7_ I first learned about the Hypothesis form of shrinking PRNG history from this talk by Martin Janiczek. He also implemented it in elm-test.\n\n- https://hypothesis.works/articles/compositional-shrinking/\n  David R. MacIver, the creator of Hypothesis and the inventor of the internal shrinking concept we are using.\n\n  > Hypothesis takes the “Shrinking outputs can be done by shrinking inputs” idea to its logical conclusion and has a single unified intermediate representation that all generation is based off.\n\n- https://drmaciver.github.io/papers/reduction-via-generation-preview.pdf research paper by David R. MacIver and Alastair F. Donaldson\n  > The key idea of internal reduction is to manipulate the underlying source of randomness\n  > consumed by a random generator, in order to cause the generator to produce smaller test\n  > cases automatically. The final reduced test case is constructed as if the generator had been\n  > implausibly lucky and produced a small and readable test case by chance.","ref":"readme.html#background"},{"type":"extras","title":"Installation - Decorum","doc":"The package can be installed\nby adding `decorum` to your list of dependencies in `mix.exs`:\n\n```elixir\ndef deps do\n  [\n    {:decorum, \"~> 0.1.1\"}\n  ]\nend\n```\n\nDocumentation is generated with [ExDoc](https://github.com/elixir-lang/ex_doc)\nand published on [HexDocs](https://hexdocs.pm). The docs can\nbe found at  .","ref":"readme.html#installation"},{"type":"extras","title":"ElixirConf EU 2024 - Decorum","doc":"I will be presenting a talk about internal shrinking and this library at [ElixirConf EU 2024](https://www.elixirconf.eu/talks/the-magic-of-internal-shrinking-for-property-based-testing/) in Lisbon on April 18-19.","ref":"readme.html#elixirconf-eu-2024"},{"type":"extras","title":"Implementation Notes","doc":"# Implementation Notes\n\nThese are my thoughts and notes while investigating how this library should be implemented.\nI'm including it in the repo to preserve revision history as I find answers to my own questions and evolve the design.\n\nReferences to Martin are refering to Martin Janiczek. I first learned about the Hypothesis form of shrinking from this YouTube video https://youtu.be/WE5bmt0zBxg?si=fyg6R_O3iRrbuc7_ He also implemented this in elm-test.\n\nElixir Outlaws ep 2 has a good quote. Something like “shrinking is the real magic of property based testing.” Ep 4 also talks about prop testing and a debate around StreamData being included in Elixir core.\n\n#","ref":"notes.html"},{"type":"extras","title":"Internal representation of PRNG history. - Implementation Notes","doc":"We're currently using a list of 32-bit integers, which is the same as the Elm implementation.\n\nUnsigned Integers work best because they shrink easily by dividing by 2 or subtracting 1.\n\nWhat is the internal representation of integers in the BEAM? Small integer is 1 word (60 bits on 64-bit architecture) See https://www.erlang.org/doc/efficiency_guide/advanced.html\n\nWhat effect does size of integer stored in PRNG history have on the rand algorithm used? If only storing 32-bit ints, then it doesn't matter.\n\n- [ ] Should we use a bytes as is done in Hypothesis?\n      No matter what size ints we store, when a larger value is requested, we need to use more than one of them.\n\nUsing smaller integers might make it more important to label chunks of random history the way hypothesis does. I couldn’t find the equivalent in the Elm implementation. Martin confirmed that it's not in the Elm code.\n\n#","ref":"notes.html#internal-representation-of-prng-history"},{"type":"extras","title":"How is `:rand` thread safe when running ExUnit async true? - Implementation Notes","doc":"We sidestep this question by remembering `:rand.state`.\nHow does that relate to the given seed?\nFrom https://hexdocs.pm/ex_unit/1.15.4/ExUnit.html#configure/1-options\n\n> `:seed` - an integer seed value to randomize the test suite. This seed is also mixed with the test module and name to create a new unique seed on every test, which is automatically fed into the :rand module. This provides randomness between tests, but predictable and reproducible results. A :seed of 0 will disable randomization and the tests in each file will always run in the order that they were defined in;\n\nGenerators in StreamData create functions that take a seed (it’s called seed,\nbut it’s really `:rand.state()`) and keeps track of the next state. Fuzzers in Elm Test take a prng parameter. (see rollDice in https://github.com/elm-explorations/test/blob/master/src/Fuzz.elm)\nThe Random prng in Elm also keeps track of the next seed. -- Maybe that's a key ingredient --\n**it doesn't matter what calls to `:rand` are made in other threads if our PRNG is keeping the next seed that it will use.**\n\nPRNG history only needs to be preserved for a single test iteration. In `check_all`, we don't need the history for the whole 100 test runs. We only need it for one that fails a test.\nHowever, we do need to use a new seed for each run. How do other implementations handle this? https://github.com/elm-explorations/test/blob/master/src/Test/Fuzz.elm has a `stepSeed` function that gets the seed for the next run.\nWe can use `:rand.jump()` for the same purpose.\n\nIt's important to not specify a new seed so that we're based on the one ExUnit sets up automatically.\n\n#","ref":"notes.html#how-is-rand-thread-safe-when-running-exunit-async-true"},{"type":"extras","title":"TODO: - Implementation Notes","doc":"- [x] On prng playback need to distinguish between no history and getting to the end of the history. Tag as :random or :hardcoded ?\n      With no history it needs to use the same seed as ExUnit, which happens automatically because we start with a call to `:rand.jump()`\n\n- [x] Simple history replay should be easy to test.\n\n- [x] Hanlde getting to the end of prng history. Change `next/1` to `next!/1` and raise EmptyHistoryError.\n\n- [x] When validating a shrunken history need to distinguish between running out of numbers and no longer failing the test.\n\n- [x] Remove `prng` parameter from `check_all` because we need a new one for each test run. Or make it optional.\n\n- [x] Catch errors raised by `body_fn` so we can capture PRNG history and enter shrinking cycle.\n\n- [ ] Format raised error message to include generated values and shrinking statistics.\n      Use Telemetry for metrics and stats?\n\n- [x] Include `value` field in `PropertyError`.\n\n- [ ] How do we get a meaningful stacktrace? Does that even matter?\n\n- [x] Get end to end with shrinking working with single integer generator.\n\n- [x] Add `list_of` with shrinking on a list of integers. Use \"list is sorted\" as the property which should shrink to `[1,0]`.\n\n- [x] Change `History.shrink_length/1` to remove one item at a time and fix the resulting error and occasional timeout.\n\n- [x] Keep track of seen histories to avoid trying them again. (No longer needed after refactoring into `Shrinker` module?)\n\n- [x] Try a new implementation of shrinking. Create multiple histories from a given history. Test all of them against test_fn.\n      Keep best (shortlex smallest) that still fails the test and re-start the shrinking process with that one as the input.\n      Copy more of the elm-test implementation. Create a `Shrinker` module.\n\n- [x] Put raw chunk manipulation functions in `History` and test them.\n\n- [x] Implement binary search for finding smaller interesting values within the PRNG history.\n\n- [ ] Make a generic version of binary search?\n\n- [x] Change `list_of` to have some maximum list size. Adjust probability as it gets closer to the max? See https://github.com/elm-explorations/test/blob/9669a27d84fc29175364c7a60d5d700771a2801e/src/Fuzz.elm#L678\n\n- [ ] Add the concept of generation size and re-sizing from StreamData?\n\n- [ ] Store `length` in `History` struct? This would make some operations more efficient.\n\n- [ ] Add support for generating integers larger than the internal representation of the PRNG history which is currently a 32-bit integer. This requires consuming more than one value. The `next` funciton probably needs a byte_count parameter.\n\n- [x] Implement Enumerable protocol for Decorum struct.\n\n- [ ] Implement float generator by copying Elm/Hypothesis implementation. How does it optimize for shrinking? (I initially guessed that it simplified towards 1.0 instead of towards zero, but that wouldn’t produce simpler fractions.) See https://github.com/HypothesisWorks/hypothesis/blob/d55849df92d01a25364aa21a1adb310ee0a3a390/hypothesis-python/src/hypothesis/internal/conjecture/floats.py which was linked to from https://github.com/elm-explorations/test/blob/master/src/Fuzz/Float.elm\n\n- [x] Run the test body N times looking for initial failures. N is how many items we take from the generator. N should be configurable, but start with 100.\n\n- [ ] Add configuration option for how many times to run the test body.\n\n- [x] Only feed used history into next round of shrinking? Discard unused values at the end of history.\n      Currently, this doesn't work with the implementation of binary search, but it does work for the shrinking\n      by chunks.\n\n- [ ] Implement other basic generators such as `string`, `boolean`, etc.\n\n- [x] Add a `zip/1` function that takes a list of generators and emits a tuple with each of their values.\n      It's essentially the same as `Enum.zip/1` but for Decorum generators.\n      It looks like StreamData has a generator named `tuple` which does this with a tuple of generators as its input.\n\n- [ ] Rename Prng module to Random?\n      I don’t love the name Prng.\n      Maybe flatten the structure while keeping `random/0` and `hardcoded/1` constructor functions.\n\n- [x] Add `filter/2` function that takes a generator and a predicate and filters out values that don't match the predicate.\n\n- [ ] Add `property` and `check all` macros. Others?\n\n- [ ] Run the shrinking challenges (https://github.com/jlink/shrinking-challenge)\n\n- [x] Publish to Hex.pm\n\n- [x] Clean up docs. The `Decorum` and maybe `Prng` modules are the only ones that need to show up in the docs.\n\n- [ ] add `mix dialyzer` to GitHub action see https://github.com/jeremyjh/dialyxir/blob/master/docs/github_actions.md\n\n- [ ] Create seperate doc sections of functions within `Decorum` module: helpers, property testing, and generators.\n\n- [ ] Should generators be a behavior? The generate function is a good use case for a callback. Does that effect the Enumerable implementation?\n\n#","ref":"notes.html#todo"},{"type":"extras","title":"How do we make generators composible? - Implementation Notes","doc":"Users should be able to create new generators based on the library generators.\n\nUsing functions such as `map/2` and `and_then/2`, new generators can be easily based on existing generators.\n`map/2` is for a simple function over generated values while `and_then/2` is for running a generator based on a generated value.\n\n#","ref":"notes.html#how-do-we-make-generators-composible"},{"type":"extras","title":"What happens when a property uses more than one generator? - Implementation Notes","doc":"Use the `zip/1` function for properties that use more than one generator.\n\n#","ref":"notes.html#what-happens-when-a-property-uses-more-than-one-generator"},{"type":"extras","title":"How long of lists should the list generator produce? - Implementation Notes","doc":"**StreamData gives lists up to generation size.**\nPropEr also uses an internal increasing size parameter. The sized function in PropCheck is used to get the current size parameter. In StreamData sized is a macro and the scale function is used to add a multiple of the size.\n\nWhat about the biased coin flip for choosing another item? The Elm implementation uses it.\nI don’t understand how the shrinker would know that the pair goes together. How could this relate to using the size parameter for affecting the length of the generated lists?\nWe could use size as a limit on the length, or we could change the weight of the coin flip as we get closer to size.\n\n#","ref":"notes.html#how-long-of-lists-should-the-list-generator-produce"},{"type":"extras","title":"What is a Generator? - Implementation Notes","doc":"A function that takes in a Prng struct (and a size?) and returns the next value and an updated Prng struct. Implementing a stream doesn’t give the updated prng struct from which to get the history. But outside of running properties, we don't need it to do that.\nIt could behave like a Stream by default and internally to `check_all` the state could be tracked. The generator function is essentially the same as `next_fun` used by `Stream.unfold`.\n\n#","ref":"notes.html#what-is-a-generator"},{"type":"extras","title":"What is a Shrinker? - Implementation Notes","doc":"A function that takes a PRNG history, a generator, and a test function and searches through a set of shortlex smaller histories.\n\nIterate through those histories until:\n\n- A. we find one that still fails the test\n- B. We don't find any that still fail the test after a full shrinking pass.\n- C. We try some maximum number of times.\n\nHow does it find possible smaller histories? By using some set of strategies or passes?\n\nWhen it finds a valid smaller history, then starts over with that one.\n\nIndividually shrink integers using binary search.\n\nKeep track of histories that have been used/seen to avoid retrying them. After refactoring into `Shrinker` module, this was not necessary.\n\nDo we shrink the first value and then shrink the rest of the history? That doesn't really work. Some shrinking needs to operate on later values only.\n\nStoring larger integers might make shrinking less efficient because it takes longer to reach low values. Binary search solves this issue.\n\nIs the process of rerunning the test and trying further shrinking similar to genetic algorithms? Not really. I do think it's similar, but I didn't find any value in trying to implement it that way over just copying how other libraries based on Hypothesis work.\n\nFor a given test, shrinking shouldn’t need to be parallelized.\n\nTest functions passed to `StreamData.check_all` do not raise exceptions. They return `{:ok, map()}` or `{:error, map()}`.\nIn the error tuple, the map contains details about original and shrunk failures.\nAlso, the private `check_all` function is recursive when the test function passes.\n\n#","ref":"notes.html#what-is-a-shrinker"},{"type":"extras","title":"Is there a better syntax for defining property tests in Elixir? - Implementation Notes","doc":"**Prefer StreamData syntax for macros like `check all`** instead of PropCheck (and PropEr) syntax of `forall`.\n\nBody of `check all` uses asserts. Body of `forall` returns boolean. Using asserts seems more like what a user of ExUnit would be familiar with.\n\nIs there a way to do it inside the test macro instead of using a property macro? Or just making the property macro the only thing that’s needed? Why require a macro inside the body of another macro?\nFrom https://github.com/whatyouhide/stream_data/blob/main/lib/ex_unit_properties.ex it looks like\n**the property macro is a convenience for marking tests as properties.**\n\n#","ref":"notes.html#is-there-a-better-syntax-for-defining-property-tests-in-elixir"},{"type":"extras","title":"Other thoughts or questions - Implementation Notes","doc":"In Elm implementation Fuzz.elm, why does forcedChoice need to consume a random number? How is that different from constant?\nMartin confirmed that throwing an error instead of just accepting it may not be right.\n\nI considered setting up property test macros and using them to build up tests for the system, but I'm not sure that would help.\nJust use `check_all` directly without the macros. Add macros later.\n\nHow important is it for `Decorum.uniform_integer/1` to produce uniformly random values?\nI tried the code from https://rosettacode.org/wiki/Verify_distribution_uniformity/Chi-squared_test and its `chi2IsUniform/2` function returned false for all the examples I ran.\nThe `chi2Probability/2` results were around `1.69e-13` when they were expected to be greater than `0.05`.","ref":"notes.html#other-thoughts-or-questions"}],"content_type":"text/markdown"}